
package ch.ethz.ssh2;

import java.io.IOException;

import ch.ethz.ssh2.auth.AuthenticationManager;
import ch.ethz.ssh2.channel.ChannelManager;
import ch.ethz.ssh2.crypto.CryptoWishList;
import ch.ethz.ssh2.crypto.cipher.BlockCipherFactory;
import ch.ethz.ssh2.crypto.digest.MAC;
import ch.ethz.ssh2.transport.TransportManager;

/**
 * A <code>Connection</code> is used to establish an encrypted TCP/IP
 * connection to a SSH2 server.<br>
 * <br>
 * Typically, one
 * <ul>
 * <li>Creates a <code>Connection</code> object.</li>
 * <li>Calls the <code>connect</code> method.</li>
 * <li>Calls one of the authentication methods.</li>
 * <li>Calls one or several times the <code>openSession</code>,
 * <code>createPortForwarder</code> or similar methods.</li>
 * <li>Finally, one can close the connection with the <code>close</code>
 * method.</li>
 * </ul>
 * 
 * @author Christian Plattner, plattner@inf.ethz.ch
 * @version $Id: Connection.java,v 1.7 2005/07/04 13:58:41 cplattne Exp $
 */

public class Connection
{
	private AuthenticationManager am;
	private boolean authenticated = false;

	private ChannelManager cm;

	private CryptoWishList cryptoWishList = new CryptoWishList();
	private DHGexParameters dhgexpara = new DHGexParameters();

	private String hostname;
	private int port;

	private TransportManager tm;

	public final static String VERSION = "$id$";
	public final static String identification = "Ganymed1.206";

	/**
	 * Same as <code>Connection(hostname, 22)</code<. 
	 * 
	 * @param hostname the hostname of the SSH2 server.
	 */
	public Connection(String hostname)
	{
		this(hostname, 22);
	}

	/**
	 * Prepares a fresh <code>Connection</code> object which can then be used
	 * to establish a connection to the specified SSH2 server.
	 * 
	 * @param hostname
	 *            the host where we laster want to connect.
	 * @param port
	 *            port on the server, normally 22.
	 */
	public Connection(String hostname, int port)
	{
		this.hostname = hostname;
		this.port = port;
	}

	/**
	 * After a successful connect, one has to authenticate oneself. This method
	 * is based on DSA (it uses DSA to sign a challenge sent by the server). On
	 * success, <code>true</code> is returned. If the server does not accept
	 * the request, <code>false</code> is returned and one can retry either by
	 * using this or any other authentication method. At the moment, only
	 * unencrypted and 3DES encrypted DSA private keys are supported. TODO: DES,
	 * AES-128, AES-192 and AES-256.
	 * 
	 * @param user
	 *            A <code>String</code> holding the username.
	 * @param pem
	 *            A <code>String</code> containing the DSA private key of the
	 *            user in OpenSSH key format (PEM, you can't miss the
	 *            "-----BEGIN DSA PRIVATE KEY-----" tag). The string may contain
	 *            linefeeds.
	 * @param password
	 *            If the PEM string is 3DES encrypted ("DES-EDE3-CBC"), then you
	 *            must specify the password. Otherwise, this argument will be
	 *            ignored and can be set to <code>null</code>.
	 * 
	 * @return whether the connection is now authenticated.
	 * @throws IOException
	 */
	public synchronized boolean authenticateWithDSA(String user, String pem, String password) throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Connection is not established!");

		if (authenticated)
			throw new IllegalStateException("Connection is already authenticated!");

		if (am == null)
			am = new AuthenticationManager(tm);

		if (cm == null)
			cm = new ChannelManager(tm);

		authenticated = am.authenticateDSA(user, pem, password);

		return authenticated;
	}

	/**
	 * After a successfull connect, one has to authenticate oneself. This method
	 * sends username and password to the server. On success, <code>true</code>
	 * is returned. If the server does not accept the username and password,
	 * <code>false</code> is returned and one can retry either by using this
	 * or any other authentication method.
	 * 
	 * @param user
	 * @param pass
	 * @return if the connection is now authenticated.
	 * @throws IOException
	 */
	public synchronized boolean authenticateWithPassword(String user, String pass) throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Connection is not established!");

		if (authenticated)
			throw new IllegalStateException("Connection is already authenticated!");

		if (am == null)
			am = new AuthenticationManager(tm);

		if (cm == null)
			cm = new ChannelManager(tm);

		authenticated = am.authenticatePassword(user, pass);

		return authenticated;
	}

	/**
	 * Close the connection to the SSH2 server. All assigned sessions will be
	 * closed, too. Can be called at any time. Don't forget to call this once
	 * you don't need a connection anymore - otherwise the receive thread will
	 * run forever.
	 */
	public synchronized void close()
	{
		if (tm != null)
		{
			tm.close("Closed due to user request.");
			tm = null;
		}
		am = null;
		cm = null;
		authenticated = false;
	}

	/**
	 * Same as <code>connect(null)</code>.
	 * 
	 * @return see comments for the <code>connect(null)</code> method.
	 * @throws IOException
	 */
	public synchronized ConnectionInfo connect() throws IOException
	{
		return connect(null);
	}

	/**
	 * Connect to the SSH2 server and, as soon as the server has presented its
	 * host key, use the <code>verifyServerHostKey</code> method of the
	 * <code>verifier</code> to ask for permission to proceed. Only correctly
	 * signed host keys will be passed to the <code>verifier</code>. If
	 * <code>verifier</code> is <code>null</code>, then any correctly
	 * signed host key will be accepted - this is NOT recommended, since it
	 * makes man-in-the-middle attackes VERY easy.<br>
	 * <br>
	 * Note: as long as the key exchange of the underlying connection has not
	 * been completed, this call will block.
	 * 
	 * @param verifier
	 *            An object that implements the
	 *            <code>ServerHostKeyVerifier</code> interface.
	 * @return A <code>ConnectionInfo</code> object containg the details of
	 *         the established connection.
	 * @throws IOException
	 *             If any problem occurs, e.g., the server's host key is not
	 *             accepted by the <code>verifier</code>.
	 */
	public synchronized ConnectionInfo connect(ServerHostKeyVerifier verifier) throws IOException
	{
		if (tm != null)
			throw new IOException("Connection is already in connected state!");

		try
		{
			tm = new TransportManager(hostname, port);
			tm.initialize(cryptoWishList, verifier, dhgexpara);

			/* Wait until first KEX has finished */

			return tm.getConnectionInfo(1);
		}
		catch (IOException e1)
		{
			close();
			throw new IOException(e1.getMessage());
		}
	}

	/**
	 * Force an asynchronous key re-exchange (the call does not block). The
	 * latest values set for MAC, Cipher and DH group exchange parameters will
	 * be used. If a key exchange is currently in progress, then this method has
	 * the only effect that the so far specified parameters will be used for the
	 * next (server driven) key exchange. NOTE: This implementation will never
	 * start a key exchange (other than the initial one) unless you or the SSH2
	 * server ask for it.
	 * 
	 * @throws IOException
	 *             In case of any failure behind the scenes.
	 */
	public synchronized void forceKeyExchange() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("You need to establish a connection first.");

		tm.forceKeyExchange(cryptoWishList, dhgexpara);
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @return The list of supported cipher algorithms by this implementation.
	 */
	public synchronized String[] getAvailableCiphers()
	{
		return BlockCipherFactory.getCipherList();
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @return The list of supported MAC algorthims by this implementation.
	 */
	public synchronized String[] getAvailableMACs()
	{
		return MAC.getMacList();
	}

	/**
	 * Returns a <code>ConnectionInfo</code> object containing the details of
	 * the connection. Can be called as soon as the connection has been
	 * established (successfully connected).
	 * 
	 * @return A <code>ConnectionInfo</code> object.
	 * @throws IOException
	 *             In case of any failure behind the scenes.
	 */
	public synchronized ConnectionInfo getConnectionInfo() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException(
					"Cannot get details of connection, you need to establish a connection first.");
		return tm.getConnectionInfo(1);
	}

	/**
	 * Open a new session on this connection. Works only after one has passed
	 * successfully the authentication step. One can open as many concurrent
	 * sessions as one wants.
	 * 
	 * @return A <code>Session</code> object.
	 * @throws IOException
	 */
	public synchronized Session openSession() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot open session, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot open session, connection is not authenticated.");

		return new Session(cm);
	}

	// /**
	// * Open a new SFTP client session on this connection. Works only after one
	// * has passed successfully the authentication step. One can open as many
	// * concurrent sessions as one wants.
	// *
	// * @return A <code>Session</code> object.
	// * @throws IOException
	// */
	// public synchronized SFTPClient openSFTP() throws IOException
	// {
	// if (tm == null)
	// throw new IllegalStateException("Cannot open sftp session, you need to
	// establish a connection first.");
	//
	// if (!authenticated)
	// throw new IllegalStateException("Cannot open sftp session, connection is
	// not authenticated.");
	//
	// return new SFTPClient(cm);
	// }

	/**
	 * Create a very basic <code>SCPClient</code> that can be used to copy
	 * files from/to the SSH2 server.
	 * 
	 * @return A <code>SCPClient</code> object.
	 * @throws IOException
	 */
	public synchronized SCPClient createSCPClient() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot create SCP client, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot create SCP client, connection is not authenticated.");

		return new SCPClient(this);
	}

	/**
	 * Creates a new <code>LocalPortForwarder</code>. A
	 * <code>LocalPortForwarder</code> forwards TCP/IP connections to a local
	 * port via the secure tunnel to another host (which may or may not be
	 * identical to the remote SSH2 server).
	 * 
	 * @return A <code>LocalPortForwarder</code> object.
	 * @throws IOException
	 */
	public synchronized LocalPortForwarder createLocalPortForwarder(int local_port, String host_to_connect,
			int port_to_connect) throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot forward ports, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot forward ports, connection is not authenticated.");

		return new LocalPortForwarder(cm, local_port, host_to_connect, port_to_connect);
	}

	/**
	 * Creates a new <code>LocalStreamForwarder</code>. A
	 * <code>LocalStreamForwarder</code> forwards an Input- and Outputstream
	 * pair via the secure tunnel to another host (which may or may not be
	 * identical to the remote SSH2 server).
	 * 
	 * @return A <code>LocalStreamForwarder</code> object.
	 * @throws IOException
	 */
	public synchronized LocalStreamForwarder createLocalStreamForwarder(String host_to_connect, int port_to_connect)
			throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot forward, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot forward, connection is not authenticated.");

		return new LocalStreamForwarder(cm, host_to_connect, port_to_connect);
	}

	/**
	 * Sets the parameters for the diffie-hellman group exchange. Unless you
	 * know what you are doing, you will never need this. Default values are
	 * defined in the <code>DHGexParameters</code> class.
	 * 
	 * @param dgp
	 *            non null.
	 */
	public synchronized void setDHGexParameters(DHGexParameters dgp)
	{
		if (dgp == null)
			throw new IllegalArgumentException();

		dhgexpara = dgp;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param ciphers
	 */
	public synchronized void setClient2ServerCiphers(String[] ciphers)
	{
		if (ciphers == null)
			throw new IllegalArgumentException();
		BlockCipherFactory.checkCipherList(ciphers);
		cryptoWishList.c2s_enc_algos = ciphers;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param macs
	 */
	public synchronized void setClient2ServerMACs(String[] macs)
	{
		if (macs == null)
			throw new IllegalArgumentException();
		MAC.checkMacList(macs);
		cryptoWishList.c2s_mac_algos = macs;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param ciphers
	 */
	public synchronized void setServer2ClientCiphers(String[] ciphers)
	{
		if (ciphers == null)
			throw new IllegalArgumentException();
		BlockCipherFactory.checkCipherList(ciphers);
		cryptoWishList.s2c_enc_algos = ciphers;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param macs
	 */
	public synchronized void setServer2ClientMACs(String[] macs)
	{
		if (macs == null)
			throw new IllegalArgumentException();
		MAC.checkMacList(macs);
		cryptoWishList.s2c_mac_algos = macs;
	}
}
